@@ -16,7 +16,7 @@ const fixtures = path.resolve(__dirname, 'fixtures')
 const tars = path.resolve(fixtures, 'tars')
 const parses = path.resolve(fixtures, 'parse')
 const unpackdir = path.resolve(fixtures, 'unpack')
-const {promisify} = require('util')
+const { promisify } = require('util')
 const rimraf = promisify(require('rimraf'))
 const mkdirp = require('mkdirp')
 const mutateFS = require('mutate-fs')
@@ -98,8 +98,9 @@ t.test('basic file unpack tests', t => {
         const expect = cases[tarfile]
         Object.keys(expect).forEach(file => {
           const f = path.resolve(dir, file)
-          if (isWindows && isLongFile(file))
+          if (isWindows && isLongFile(file)) {
             return
+          }
           t.equal(fs.readFileSync(f, 'utf8'), expect[file], file)
         })
         t.end()
@@ -189,16 +190,8 @@ t.test('links!', t => {
     t.end()
   }
   const checkForStrip3 = t => {
-    t.ok(fs.lstatSync(dir + '/3').isDirectory())
-    let err = null
-    try {
-      fs.lstatSync(dir + '/3/hardlink-3')
-    } catch (e) {
-      err = e
-    }
-    // can't be extracted because we've passed it in the tar
-    // (specially crafted tar for this not to work)
-    t.equal(err.code, 'ENOENT')
+    // strips the linkpath entirely, so the link doesn't get extracted.
+    t.throws(() => fs.lstatSync(dir + '/3'), { code: 'ENOENT' })
     t.end()
   }
 
@@ -647,11 +640,11 @@ t.test('unsupported entries', t => {
     const c = 'TAR_ENTRY_UNSUPPORTED'
     const expect = [
       [c, 'unsupported entry type: CharacterDevice', {
-        entry: { path: 'dev/random' }}],
+        entry: { path: 'dev/random' } }],
       [c, 'unsupported entry type: BlockDevice', {
-        entry: { path: 'dev/hd0' }}],
+        entry: { path: 'dev/hd0' } }],
       [c, 'unsupported entry type: FIFO', {
-        entry: { path: 'dev/fifo0' }}],
+        entry: { path: 'dev/fifo0' } }],
     ]
     u.on('close', _ => {
       t.equal(fs.readdirSync(dir).length, 0)
@@ -807,7 +800,7 @@ t.test('absolute paths', t => {
   t.ok(path.isAbsolute(extraAbsolute))
   t.ok(path.isAbsolute(absolute))
   const parsed = path.parse(absolute)
-  const relative = absolute.substr(parsed.root.length)
+  const relative = absolute.slice(parsed.root.length)
   t.notOk(path.isAbsolute(relative))
 
   const data = makeTar([
@@ -1027,13 +1020,15 @@ t.test('fail all stats', t => {
   const mutate = () => {
     fs.stat = fs.lstat = fs.fstat = (...args) => {
       // don't fail statting the cwd, or we get different errors
-      if (normPath(args[0]) === dir)
+      if (normPath(args[0]) === dir) {
         return lstat(dir, args.pop())
+      }
       process.nextTick(() => args.pop()(poop))
     }
     fs.statSync = fs.lstatSync = fs.fstatSync = (...args) => {
-      if (normPath(args[0]) === dir)
+      if (normPath(args[0]) === dir) {
         return lstatSync(dir)
+      }
       throw poop
     }
   }
@@ -2016,7 +2011,7 @@ t.test('dont use explicit chmod if noChmod flag set', t => {
 
   return t.test('sync', t => {
     mkdirp.sync(basedir)
-    const unpack = new Unpack.Sync({ cwd: basedir, noChmod: true})
+    const unpack = new Unpack.Sync({ cwd: basedir, noChmod: true })
     unpack.end(data)
     check(t)
   })
@@ -2628,7 +2623,7 @@ t.test('trying to unpack a non-zlib gzip file should fail', t => {
     const skip = !/^v([0-9]|1[0-3])\./.test(process.version) ? false
       : 'node prior to v14 did not raise sync zlib errors properly'
     t.throws(() => new UnpackSync(opts).end(dataGzip),
-      expect, 'sync throws', {skip})
+      expect, 'sync throws', { skip })
   })
 
   t.test('bad archive if no gzip', t => {
@@ -2890,7 +2885,7 @@ t.test('close fd when error writing', t => {
   t.teardown(mutateFS.fail('write', new Error('nope')))
   const CLOSES = []
   const OPENS = {}
-  const {open} = require('fs')
+  const { open } = require('fs')
   t.teardown(() => fs.open = open)
   fs.open = (...args) => {
     const cb = args.pop()
@@ -2912,8 +2907,9 @@ t.test('close fd when error writing', t => {
     onwarn: (code, msg) => WARNINGS.push([code, msg]),
   })
   unpack.on('end', () => {
-    for (const [path, fd] of Object.entries(OPENS))
+    for (const [path, fd] of Object.entries(OPENS)) {
       t.equal(CLOSES.includes(fd), true, 'closed fd for ' + path)
+    }
     t.end()
   })
   unpack.end(data)
@@ -2942,7 +2938,7 @@ t.test('close fd when error setting mtime', t => {
   t.teardown(mutateFS.fail('utimes', new Error('nooooope')))
   const CLOSES = []
   const OPENS = {}
-  const {open} = require('fs')
+  const { open } = require('fs')
   t.teardown(() => fs.open = open)
   fs.open = (...args) => {
     const cb = args.pop()
@@ -2964,8 +2960,9 @@ t.test('close fd when error setting mtime', t => {
     onwarn: (code, msg) => WARNINGS.push([code, msg]),
   })
   unpack.on('end', () => {
-    for (const [path, fd] of Object.entries(OPENS))
+    for (const [path, fd] of Object.entries(OPENS)) {
       t.equal(CLOSES.includes(fd), true, 'closed fd for ' + path)
+    }
     t.end()
   })
   unpack.end(data)
@@ -3009,3 +3006,230 @@ t.test('do not hang on large files that fail to open()', t => {
     })
   })
 })
+
+t.test('dirCache pruning unicode normalized collisions', {
+  skip: isWindows && 'symlinks not fully supported',
+}, t => {
+  const data = makeTar([
+    {
+      type: 'Directory',
+      path: 'foo',
+    },
+    {
+      type: 'File',
+      path: 'foo/bar',
+      size: 1,
+    },
+    'x',
+    {
+      type: 'Directory',
+      // café
+      path: Buffer.from([0x63, 0x61, 0x66, 0xc3, 0xa9]).toString(),
+    },
+    {
+      type: 'SymbolicLink',
+      // cafe with a `
+      path: Buffer.from([0x63, 0x61, 0x66, 0x65, 0xcc, 0x81]).toString(),
+      linkpath: 'foo',
+    },
+    {
+      type: 'Directory',
+      path: 'foo',
+    },
+    {
+      type: 'File',
+      path: Buffer.from([0x63, 0x61, 0x66, 0xc3, 0xa9]).toString() + '/bar',
+      size: 1,
+    },
+    'y',
+    '',
+    '',
+  ])
+
+  const check = (path, dirCache, t) => {
+    path = path.replace(/\\/g, '/')
+    t.strictSame([...dirCache.entries()][0], [`${path}/foo`, true])
+    t.equal(fs.readFileSync(path + '/foo/bar', 'utf8'), 'x')
+    t.end()
+  }
+
+  t.test('sync', t => {
+    const path = t.testdir()
+    const dirCache = new Map()
+    new UnpackSync({ cwd: path, dirCache }).end(data)
+    check(path, dirCache, t)
+  })
+  t.test('async', t => {
+    const path = t.testdir()
+    const dirCache = new Map()
+    new Unpack({ cwd: path, dirCache })
+      .on('close', () => check(path, dirCache, t))
+      .end(data)
+  })
+
+  t.end()
+})
+
+t.test('dircache prune all on windows when symlink encountered', t => {
+  if (process.platform !== 'win32') {
+    process.env.TESTING_TAR_FAKE_PLATFORM = 'win32'
+    t.teardown(() => {
+      delete process.env.TESTING_TAR_FAKE_PLATFORM
+    })
+  }
+  const symlinks = []
+  const Unpack = t.mock('../lib/unpack.js', {
+    fs: {
+      ...fs,
+      symlink: (target, dest, cb) => {
+        symlinks.push(['async', target, dest])
+        process.nextTick(cb)
+      },
+      symlinkSync: (target, dest) => symlinks.push(['sync', target, dest]),
+    },
+  })
+  const UnpackSync = Unpack.Sync
+
+  const data = makeTar([
+    {
+      type: 'Directory',
+      path: 'foo',
+    },
+    {
+      type: 'File',
+      path: 'foo/bar',
+      size: 1,
+    },
+    'x',
+    {
+      type: 'Directory',
+      // café
+      path: Buffer.from([0x63, 0x61, 0x66, 0xc3, 0xa9]).toString(),
+    },
+    {
+      type: 'SymbolicLink',
+      // cafe with a `
+      path: Buffer.from([0x63, 0x61, 0x66, 0x65, 0xcc, 0x81]).toString(),
+      linkpath: 'safe/actually/but/cannot/be/too/careful',
+    },
+    {
+      type: 'File',
+      path: 'bar/baz',
+      size: 1,
+    },
+    'z',
+    '',
+    '',
+  ])
+
+  const check = (path, dirCache, t) => {
+    // symlink blew away all dirCache entries before it
+    path = path.replace(/\\/g, '/')
+    t.strictSame([...dirCache.entries()], [
+      [`${path}/bar`, true],
+    ])
+    t.equal(fs.readFileSync(`${path}/foo/bar`, 'utf8'), 'x')
+    t.equal(fs.readFileSync(`${path}/bar/baz`, 'utf8'), 'z')
+    t.end()
+  }
+
+  t.test('sync', t => {
+    const path = t.testdir()
+    const dirCache = new Map()
+    new UnpackSync({ cwd: path, dirCache }).end(data)
+    check(path, dirCache, t)
+  })
+
+  t.test('async', t => {
+    const path = t.testdir()
+    const dirCache = new Map()
+    new Unpack({ cwd: path, dirCache })
+      .on('close', () => check(path, dirCache, t))
+      .end(data)
+  })
+
+  t.end()
+})
+
+t.test('recognize C:.. as a dot path part', t => {
+  if (process.platform !== 'win32') {
+    process.env.TESTING_TAR_FAKE_PLATFORM = 'win32'
+    t.teardown(() => {
+      delete process.env.TESTING_TAR_FAKE_PLATFORM
+    })
+  }
+  const Unpack = t.mock('../lib/unpack.js', {
+    path: {
+      ...path.win32,
+      win32: path.win32,
+      posix: path.posix,
+    },
+  })
+  const UnpackSync = Unpack.Sync
+
+  const data = makeTar([
+    {
+      type: 'File',
+      path: 'C:../x/y/z',
+      size: 1,
+    },
+    'z',
+    {
+      type: 'File',
+      path: 'x:..\\y\\z',
+      size: 1,
+    },
+    'x',
+    {
+      type: 'File',
+      path: 'Y:foo',
+      size: 1,
+    },
+    'y',
+    '',
+    '',
+  ])
+
+  const check = (path, warnings, t) => {
+    t.equal(fs.readFileSync(`${path}/foo`, 'utf8'), 'y')
+    t.strictSame(warnings, [
+      [
+        'TAR_ENTRY_ERROR',
+        "path contains '..'",
+        'C:../x/y/z',
+        'C:../x/y/z',
+      ],
+      ['TAR_ENTRY_ERROR', "path contains '..'", 'x:../y/z', 'x:../y/z'],
+      [
+        'TAR_ENTRY_INFO',
+        'stripping Y: from absolute path',
+        'Y:foo',
+        'foo',
+      ],
+    ])
+    t.end()
+  }
+
+  t.test('async', t => {
+    const warnings = []
+    const path = t.testdir()
+    new Unpack({
+      cwd: path,
+      onwarn: (c, w, { entry, path }) => warnings.push([c, w, path, entry.path]),
+    })
+      .on('close', () => check(path, warnings, t))
+      .end(data)
+  })
+
+  t.test('sync', t => {
+    const warnings = []
+    const path = t.testdir()
+    new UnpackSync({
+      cwd: path,
+      onwarn: (c, w, { entry, path }) => warnings.push([c, w, path, entry.path]),
+    }).end(data)
+    check(path, warnings, t)
+  })
+
+  t.end()
+})
